unit uTags;

interface

uses
  SysUtils, Classes;

type
  TAnsiCharSet = set of AnsiChar;

procedure FindHtmlTags(const FileName: string; TagList: TStringList = nil; EntityList: TStringList = nil; HighAsciiList: TStringList = nil; const SkipAsciiHigh: TAnsiCharSet = []);
procedure FindHtmlTagsInFiles(const Mask: string; Recursive: boolean; TagList: TStringList = nil; EntityList: TStringList = nil; HighAsciiList: TStringList = nil; const SkipAsciiHigh: TAnsiCharSet = []);

implementation

uses
  uFiles;

function GetTag(TagStart, TagEnd, MaxTagEnd: PAnsiChar): string;
var n: integer;
begin
  if Integer(TagEnd) < Integer(MaxTagEnd) then
    n := Integer(TagEnd) - Integer(TagStart)
  else
    n := Integer(MaxTagEnd) - Integer(TagStart);
  SetLength(Result, n);
  Move(TagStart^, Result[1], n);
end;

procedure SetupList(List: TStringList);
begin
  if List <> nil then
    begin
    List.Sorted := True;
    end;
end;

procedure AddTagToList(const Tag: string; List: TStringList);
var i: integer;
begin
  if List <> nil then
    if Tag <> '' then
      if List.Find(Tag, i) then
        List.Objects[i] := TObject(Succ(Integer(List.Objects[i])))
      else
        List.AddObject(Tag, TObject(1));
end;

procedure FindHtmlTags(const FileName: string; TagList, EntityList, HighAsciiList: TStringList; const SkipAsciiHigh: TAnsiCharSet);

  procedure CheckHighAsciiAndInc(var P: PAnsiChar);
  var C: AnsiChar;
  begin
    if HighAsciiList <> nil then
      if P^ >= #128 then
        if not (P^ in SkipAsciiHigh) then
          begin
          C := P^;
          AddTagToList(C, HighAsciiList);
          end;
    Inc(P);
  end;

const EncodedChars_EntityName = ['a'..'z', 'A'..'Z', '0'..'9'];
      EncodedChars_CharCode = ['0'..'9', 'a'..'f', 'A'..'F'];
var Str: TMemoryStream;
    P, PEnd, TagStart: PAnsiChar;
    QuoteChar: AnsiChar;
    Tag: AnsiString;
    ValidChars: set of AnsiChar;
begin
  Str := LoadFileIntoMemory(FileName);
  try
    SetupList(TagList);
    SetupList(EntityList);
    SetupList(HighAsciiList);
    P := Str.Memory;
    PEnd := Pointer(Integer(P) + Str.Size);
    while Integer(P) < Integer(PEnd) do
      case P^ of
        '<':
          begin
            TagStart := P;
            while (Integer(P) < Integer(PEnd)) do
              case P^ of
                '>':
                  begin
                    Inc(P);
                    Break;
                  end;
                '"', '''':
                  begin
                    QuoteChar := P^;
                    Inc(P);
                    while (Integer(P) < Integer(PEnd)) and (P^ <> QuoteChar) do
                      CheckHighAsciiAndInc(P);
                    Inc(P);
                  end;
                else
                  begin
                    CheckHighAsciiAndInc(P);
                  end;
                end;
            Tag := GetTag(TagStart, P, PEnd);
            if (Length(Tag) > 1) and (Tag[2] <> '/') then
              AddTagToList(Tag, TagList);
          end;
        '&':
          begin
            TagStart := P;
            Inc(P);
            if (Integer(P) < Integer(PEnd)) then
              begin
              if P^ = '#' then
                begin
                ValidChars := EncodedChars_CharCode;
                Inc(P);
                end
              else
                ValidChars := EncodedChars_EntityName;
              while (Integer(P) < Integer(PEnd)) and (P^ in ValidChars) do
                CheckHighAsciiAndInc(P);
              if (Integer(P) < Integer(PEnd)) and (P^ = ';') then
                Inc(P);
              end;
            AddTagToList(GetTag(TagStart, P, PEnd), EntityList);
          end;
        else
          begin
            CheckHighAsciiAndInc(P);
          end;
        end;
  finally
    Str.Free;
    end;
end;

procedure FindHtmlTagsInFiles(const Mask: string; Recursive: boolean; TagList, EntityList, HighAsciiList: TStringList; const SkipAsciiHigh: TAnsiCharSet);
var SR: TSearchRec;
    Path, WildCard: string;
begin
  Path := ExtractFilePath(Mask);
  if FindFirst(Mask, faAnyFile, SR) = 0 then
    try
      repeat
        if not LongBool(SR.Attr and (faDirectory or faVolumeID)) then
          FindHtmlTags(Path + SR.Name, TagList, EntityList, HighAsciiList, SkipAsciiHigh);
      until FindNext(SR) <> 0;
    finally
      FindClose(SR);
      end;
  if Recursive then
    begin
    WildCard := ExtractFileName(Mask);
    if FindFirst(Path + '*.*', faDirectory, SR) = 0 then
      try
        repeat
          if LongBool(SR.Attr and faDirectory) then
            if (SR.Name <> '.') and (SR.Name <> '..') then
              FindHtmlTagsInFiles(Path + SR.Name + '\' + WildCard, Recursive, TagList, EntityList, HighAsciiList, SkipAsciiHigh);
        until FindNext(SR) <> 0;
      finally
        FindClose(SR);
        end;
    end;
end;

end.
